# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib

from pymedphys._imports import pydicom


def filter_ct_uids(
    structure_uids,
    structure_uid_to_ct_uids,
    structure_names_by_structure_set_uid,
    structure_names_by_ct_uid,
    study_set_must_have_all_of,
    slice_at_least_one_of,
    slice_must_have,
    slice_cannot_have,
):

    if structure_uids is None:
        structure_uids = structure_uid_to_ct_uids.keys()

    filtered_ct_uids = []

    for structure_uid in structure_uids:
        ct_uids = structure_uid_to_ct_uids[structure_uid]

        structure_names_in_study_set = set(
            structure_names_by_structure_set_uid[structure_uid]
        )

        if not structure_names_in_study_set.issuperset(study_set_must_have_all_of):
            continue

        for ct_uid in ct_uids:
            try:
                structure_names_on_slice = set(structure_names_by_ct_uid[ct_uid])
            except KeyError:
                structure_names_on_slice = set([])

            if len(structure_names_on_slice.intersection(slice_at_least_one_of)) == 0:
                continue

            if not structure_names_on_slice.issuperset(slice_must_have):
                continue

            if len(structure_names_on_slice.intersection(slice_cannot_have)) != 0:
                continue

            filtered_ct_uids.append(ct_uid)

    return filtered_ct_uids


def load_names_mapping(path):
    with open(path) as f:
        name_mappings_config = json.load(f)
        names_map = name_mappings_config["names_map"]
        ignore_list = name_mappings_config["ignore_list"]

        for key in ignore_list:
            names_map[key] = None

    return names_map


def verify_all_names_have_mapping(data_path_root, structure_set_paths, names_map):
    data_path_root = pathlib.Path(data_path_root)
    raw_structure_names_cache_path = data_path_root.joinpath(
        "raw-structure-names-cache.json"
    )

    relative_structure_set_paths = {
        key: str(pathlib.Path(path).relative_to(data_path_root))
        for key, path in structure_set_paths.items()
    }

    try:
        with open(raw_structure_names_cache_path) as f:
            raw_structure_names_cache = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        raw_structure_names_cache = {
            "names_in_dicom_files": [],
            "structure_set_paths_when_run": {},
        }

    cache_valid = (
        raw_structure_names_cache["structure_set_paths_when_run"]
        == relative_structure_set_paths
    )

    if not cache_valid:
        names_in_dicom_files = set()

        for _, path in structure_set_paths.items():
            dcm = pydicom.read_file(
                path, force=True, specific_tags=["StructureSetROISequence"]
            )
            for item in dcm.StructureSetROISequence:
                names_in_dicom_files.add(item.ROIName)

        raw_structure_names_cache = {
            "names_in_dicom_files": list(names_in_dicom_files),
            "structure_set_paths_when_run": relative_structure_set_paths,
        }

        with open(raw_structure_names_cache_path, "w") as f:
            json.dump(raw_structure_names_cache, f)

    names_in_dicom_files = set(raw_structure_names_cache["names_in_dicom_files"])
    mapped_names = set(names_map.keys())

    false_mapping = mapped_names.difference(names_in_dicom_files)
    to_be_mapped = names_in_dicom_files.difference(mapped_names)

    result = {
        "Names mapped that don't exist in DICOM files": false_mapping,
        "Names within DICOM files that have not been mapped yet": to_be_mapped,
    }

    return result



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

import numpy as np

from pymedphys._dicom.create import dicom_dataset_from_dict
from pymedphys._dicom.structure import (
    Structure,
    create_contour_sequence_dict,
    pull_structure,
)

A_STRUCTURE = Structure(
    "A Structure Name",
    10,
    [
        [[0, 1, 2, 0], [2, 3, 4, 2], [5, 5, 5, 5]],
        [[0, 1, 2, 0], [2, 3, 4, 2], [6, 6, 6, 6]],
        [[0, 1, 2, 0], [2, 3, 4, 2], [7, 7, 7, 7]],
    ],
)

ANOTHER_STRUCTURE = Structure(
    "Another Structure Name",
    1,
    [
        [[5, 6, 3, 5], [7, 7, 5, 7], [1, 1, 1, 1]],
        [[5, 6, 3, 5], [7, 7, 5, 7], [2, 2, 2, 2]],
        [[5, 6, 3, 5], [7, 7, 5, 7], [3, 3, 3, 3]],
    ],
)


@pytest.mark.pydicom
def test_pull_structure():
    dicom_structure = dicom_dataset_from_dict(
        {
            "StructureSetROISequence": [
                {"ROIName": A_STRUCTURE.name, "ROINumber": A_STRUCTURE.number},
                {
                    "ROIName": ANOTHER_STRUCTURE.name,
                    "ROINumber": ANOTHER_STRUCTURE.number,
                },
            ],
            "ROIContourSequence": [
                # Sequence purposely placed in reverse order to ensure ROI
                # number is being used and not list order.
                create_contour_sequence_dict(ANOTHER_STRUCTURE),
                create_contour_sequence_dict(A_STRUCTURE),
            ],
        }
    )

    x, y, z = pull_structure(A_STRUCTURE.name, dicom_structure)

    assert np.all(x == np.array(A_STRUCTURE.coords)[:, 0, :])
    assert np.all(y == np.array(A_STRUCTURE.coords)[:, 1, :])
    assert np.all(z == np.array(A_STRUCTURE.coords)[:, 2, :])



# Copyright (c) 2014-2019 James Kerns

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions
# of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# Adapted from https://github.com/jrkerns/pylinac/tree/698254258ff4cb87812840c42b34c93ae32a4693


# pylint: disable = redefined-argument-from-local, unidiomatic-typecheck


"""Utility functions for pylinac."""

import collections.abc
import decimal
import os
import os.path as osp
import struct
import subprocess
from datetime import datetime

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom


def clear_data_files():
    """Delete all demo files, image classifiers, etc from the demo folder"""
    demo_folder = osp.join(osp.dirname(osp.dirname(__file__)), "demo_files")
    if osp.isdir(demo_folder):
        for file in os.listdir(demo_folder):
            full_file = osp.join(demo_folder, file)
            if osp.isfile(full_file):
                os.remove(full_file)
    print("Pylinac data files cleared.")


def assign2machine(source_file: str, machine_file: str):
    """Assign a DICOM RT Plan file to a specific machine. The source file is overwritten to contain
    the machine of the machine file.

    Parameters
    ----------
    source_file : str
        Path to the DICOM RTPlan file that contains the fields/plan desired
        (e.g. a Winston Lutz set of fields or Varian's default PF files).
    machine_file : str
        Path to a DICOM RTPlan file that has the desired machine. This is easily obtained from pushing a plan from the TPS
        for that specific machine. The file must contain at least one valid field.
    """
    dcm_source = pydicom.dcmread(source_file)
    dcm_machine = pydicom.dcmread(machine_file)
    for beam in dcm_source.BeamSequence:
        beam.TreatmentMachineName = dcm_machine.BeamSequence[0].TreatmentMachineName
    dcm_source.save_as(source_file)


def is_close(val, target, delta=1):
    """Return whether the value is near the target value(s).

    Parameters
    ----------
    val : number
        The value being compared against.
    target : number, iterable
        If a number, the values are simply evaluated.
        If a sequence, each target is compared to ``val``.
        If any values of ``target`` are close, the comparison is considered True.

    Returns
    -------
    bool
    """
    try:
        targets = (value for value in target)
    except (AttributeError, TypeError):
        targets = [target]  # type: ignore
    for target in targets:
        if target - delta < val < target + delta:
            return True
    return False


def typed_property(name, expected_type_or_tuple_of_types):
    """Type-enforced property. Python Cookbook 9.21 (3rd ed)."""
    storage_name = "_" + name

    @property
    def prop(self):
        return getattr(self, storage_name, None)

    @prop.setter
    def prop(self, value):
        if not isinstance(value, expected_type_or_tuple_of_types):
            raise TypeError(
                f"{name} must be a {expected_type_or_tuple_of_types}. Got: {type(value)}"
            )
        setattr(self, storage_name, value)

    return prop


def simple_round(number, decimals: int = 0):
    """Round a number to the given number of decimals. Fixes small floating number errors."""
    num = int(round(number * 10 ** decimals))
    num /= 10 ** decimals
    return num


def isnumeric(obj):
    """Check whether the passed object is numeric in any sense."""
    return isinstance(obj, (int, float, decimal.Decimal, np.number))


def is_float_like(number):
    return isinstance(number, (float, np.float, np.float16, np.float32, np.float64))


def is_int_like(number):
    return isinstance(number, (int, np.int, np.int16, np.int32, np.int64, np.int8))


def is_iterable(obj):
    """Determine if an object is iterable."""
    return isinstance(obj, collections.abc.Iterable)


class Structure:
    """A simple structure that assigns the arguments to the object."""

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def update(self, **kwargs):
        self.__dict__.update(**kwargs)


def decode_binary(file, dtype, num_values=1, cursor_shift=0):
    """Read in a raw binary file and convert it to given data types.

    Parameters
    ----------
    file : file object
        The open file object.
    dtype : int, float, str
        The expected data type to return. If int or float, will return numpy array.
    num_values : int
        The expected number of dtype to return

        .. note:: This is not the same as the number of bytes.

    cursor_shift : int
        The number of bytes to move the cursor forward after decoding. This is used if there is a
        reserved section after the read-in segment.
    """
    f = file

    if dtype == str:  # if string
        output = f.read(num_values)
        if type(f) is not str:  # in py3 fc will be bytes
            output = output.decode()
        # strip the padding ("\x00")
        output = output.strip("\x00")
    elif dtype == int:
        ssize = struct.calcsize("i") * num_values
        output = np.asarray(struct.unpack("i" * num_values, f.read(ssize)))
        if len(output) == 1:
            output = int(output)
    elif dtype == float:
        ssize = struct.calcsize("f") * num_values
        output = np.asarray(struct.unpack("f" * num_values, f.read(ssize)))
        if len(output) == 1:
            output = float(output)
    else:
        raise TypeError(f"datatype '{dtype}' was not valid")

    # shift cursor if need be (e.g. if a reserved section follows)
    if cursor_shift:
        f.seek(cursor_shift, 1)
    return output


def open_path(path: str):
    """Open the specified path in the system default viewer."""

    if os.name == "darwin":
        launcher = "open"
    elif os.name == "posix":
        launcher = "xdg-open"
    elif os.name == "nt":
        launcher = "explorer"
    subprocess.call([launcher, path])


def file_exists(filename: str):
    """Check if the file exists and if it does add a timestamp"""
    if osp.exists(filename):
        filename, ext = osp.splitext(filename)
        mytime = datetime.now().strftime("%Y%m%d%H%M%S")
        filename = filename + mytime + ext
    return filename



# Copyright (C) 2020 University of New South Wales & Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from contextlib import contextmanager

import psutil


@contextmanager
def process(*args, **kwargs):
    """Provides a process running with the provided arguments, useful for CLI unit tests

    Yields
    -------
    subprocess.Popen
        The process being run
    """
    proc = subprocess.Popen(*args, **kwargs)
    try:
        yield proc
    finally:
        for child in psutil.Process(proc.pid).children(recursive=True):
            child.kill()
        proc.kill()



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imageio
import numpy as np

import pymedphys
import pymedphys._wlutz.imginterp


def test_interp_field():
    image_path = pymedphys.data_path("wlutz_image.png")
    img = imageio.imread(image_path)
    assert np.shape(img) == (1024, 1024)
    img = img[:, 1:-1]
    assert np.shape(img) == (1024, 1022)
    assert img.dtype == np.dtype("uint16")
    img = 1 - img[::-1, :] / 2 ** 16
    assert img.dtype == np.dtype("float64")

    shape = np.shape(img)
    x = np.arange(-shape[1] / 2, shape[1] / 2) / 4
    y = np.arange(-shape[0] / 2, shape[0] / 2) / 4
    xx, yy = np.meshgrid(x, y)

    field = pymedphys._wlutz.imginterp.create_interpolated_field(  # pylint:disable = protected-access
        x, y, img
    )

    assert np.all(field(xx, yy) == img)



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from copy import deepcopy

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom


@functools.lru_cache(maxsize=1)
def get_dicom_names():
    DICOM_NAMES = [item[-1] for _, item in pydicom.datadict.DicomDictionary.items()]

    return DICOM_NAMES


def add_array_to_dataset(dataset, key, value):
    if isinstance(value, np.ndarray):
        value = value.tolist()
    setattr(dataset, key, value)


def set_default_transfer_syntax(dataset):
    if dataset.is_little_endian is None:
        dataset.is_little_endian = True

    if dataset.is_implicit_VR is None:
        dataset.is_implicit_VR = True


def dicom_dataset_from_dict(input_dict: dict, template_ds=None):
    """Create a pydicom DICOM object from a dictionary"""
    if template_ds is None:
        dataset = pydicom.Dataset()
    else:
        dataset = deepcopy(template_ds)

    for key, value in input_dict.items():
        if key not in get_dicom_names():
            raise ValueError("{} is not within the DICOM dictionary.".format(key))

        if isinstance(value, dict):
            setattr(dataset, key, dicom_dataset_from_dict(value))
        elif isinstance(value, list):
            # TODO: Check for DICOM SQ type on this attribute
            if np.all([not isinstance(item, dict) for item in value]):
                add_array_to_dataset(dataset, key, value)
            elif np.all([isinstance(item, dict) for item in value]):
                setattr(dataset, key, [dicom_dataset_from_dict(item) for item in value])
            else:
                raise ValueError(
                    "{} should contain either only dictionaries, or no "
                    "dictionaries".format(key)
                )
        else:
            add_array_to_dataset(dataset, key, value)

    set_default_transfer_syntax(dataset)

    return dataset


# def structure



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import findbb, findfield, imginterp


def find_field(
    x, y, img, edge_lengths, penumbra=2, fixed_rotation=None, pylinac_tol=0.2
):
    field = imginterp.create_interpolated_field(x, y, img)
    initial_centre = findfield.get_initial_centre(x, y, img)

    field_centre, field_rotation = findfield.field_centre_and_rotation_refining(
        field,
        edge_lengths,
        penumbra,
        initial_centre,
        fixed_rotation=fixed_rotation,
        pylinac_tol=pylinac_tol,
    )

    return field, field_centre, field_rotation


def find_field_and_bb(
    x,
    y,
    img,
    edge_lengths,
    bb_diameter,
    penumbra=2,
    fixed_rotation=None,
    pylinac_tol=0.2,
):
    field, field_centre, field_rotation = find_field(
        x,
        y,
        img,
        edge_lengths,
        penumbra=penumbra,
        fixed_rotation=fixed_rotation,
        pylinac_tol=pylinac_tol,
    )

    bb_centre = findbb.optimise_bb_centre(
        field,
        bb_diameter,
        edge_lengths,
        penumbra,
        field_centre,
        field_rotation,
        pylinac_tol=pylinac_tol,
    )

    return bb_centre, field_centre, field_rotation



# pylint: disable = unused-import


from pymedphys._experimental.pseudonymisation import (
    get_default_pseudonymisation_keywords,
    is_valid_strategy_for_keywords,
    pseudonymise,
)
from pymedphys._experimental.pseudonymisation.strategy import pseudonymisation_dispatch



# Copyright (C) 2018 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable=C0103,C1801


"""Testing of a single mlc pair.
"""

import numpy as np

from pymedphys._metersetmap.metersetmap import single_mlc_pair


def test_minimal_variance_with_resolution():
    mlc_left = (-2.3, 3.1)
    mlc_right = (0, 7.7)

    x_coarse, metersetmap_coarse = single_mlc_pair(mlc_left, mlc_right, 1)
    x_fine, metersetmap_fine = single_mlc_pair(mlc_left, mlc_right, 0.01)

    reference = np.argmin(np.abs(x_fine[None, :] - x_coarse[:, None]), axis=0)

    average_metersetmap_fine = []
    for i in range(2, len(x_coarse) - 2):
        average_metersetmap_fine.append(np.mean(metersetmap_fine[reference == i]))

    average_metersetmap_fine = np.array(average_metersetmap_fine)

    assert np.allclose(average_metersetmap_fine, metersetmap_coarse[2:-2], 0.1)


def test_stationary_partial_occlusion():
    _, metersetmap = single_mlc_pair((-1, -1), (2.7, 2.7), 1)

    assert np.allclose(metersetmap, [0.5, 1, 1, 1, 0.2])


def test_large_travel():
    x, metersetmap = single_mlc_pair((-400, 400), (400, 400))

    linear = (x + 400) / 800
    linear[-1] = 0.5

    assert np.allclose(linear, metersetmap, atol=0.001)



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A toolbox for connecting to Mosaiq SQL.
"""

from contextlib import contextmanager
from getpass import getpass

from pymedphys._imports import keyring, pymssql


def execute_sql(cursor, sql_string, parameters=None):
    """Executes a given SQL string on an SQL cursor.
    """
    try:
        cursor.execute(sql_string, parameters)
    except Exception:
        print("sql_string:\n    {}\nparameters:\n    {}".format(sql_string, parameters))
        raise

    data = []

    while True:
        row = cursor.fetchone()
        if row is None:
            break

        data.append(row)

    return data


def get_username_password(
    storage_name, user_input=input, password_input=getpass, output=print
):

    user = keyring.get_password("MosaiqSQL_username", storage_name)
    password = keyring.get_password("MosaiqSQL_password", storage_name)

    if user is None or user == "":
        output(
            "Provide a user that only has `db_datareader` access to the "
            "Mosaiq database at `{}`".format(storage_name)
        )
        user = user_input()
        if user == "":
            error_message = "Username should not be blank."
            output(error_message)
            raise ValueError(error_message)
        keyring.set_password("MosaiqSQL_username", storage_name, user)

    if password is None:
        output(
            "Provide password for '{}' server and '{}' user".format(storage_name, user)
        )
        password = password_input()
        keyring.set_password("MosaiqSQL_password", storage_name, password)

    return user, password


def separate_server_port_string(sql_server_and_port):
    split_tuple = str(sql_server_and_port).split(":")
    if len(split_tuple) == 1:
        server = split_tuple[0]
        port = 1433
    elif len(split_tuple) == 2:
        server, port = split_tuple
    else:
        raise ValueError(
            "Only one : should appear in server name,"
            " and it should be used to divide hostname from port number"
        )

    return server, port


def try_connect_delete_user_if_fail(
    sql_server_and_port, user_input=input, password_input=getpass, output=print
):
    server, port = separate_server_port_string(sql_server_and_port)
    user, password = get_username_password(
        sql_server_and_port,
        user_input=user_input,
        password_input=password_input,
        output=output,
    )

    try:
        conn = pymssql.connect(server, user, password, "MOSAIQ", port=port)
    except pymssql.OperationalError as error:
        error_message = error.args[0][1]
        if error_message.startswith(b"Login failed for user"):
            output("Login failed, wiping the saved username and password.")
            try:
                keyring.delete_password("MosaiqSQL_username", sql_server_and_port)
                keyring.delete_password("MosaiqSQL_password", sql_server_and_port)
            except keyring.errors.PasswordDeleteError as error:
                pass
            output("Please try login again:")
            conn = try_connect_delete_user_if_fail(sql_server_and_port)
        else:
            output(
                "Server Input: {}, User: {}, Hostname: {}, Port: {}".format(
                    sql_server_and_port, user, server, port
                )
            )
            raise

    except Exception as error:
        output(
            "Server Input: {}, User: {}, Hostname: {}, Port: {}".format(
                sql_server_and_port, user, server, port
            )
        )
        raise

    return conn


def single_connect(
    sql_server_and_port, user_input=input, password_input=getpass, output=print
):
    """Connect to the Mosaiq server.
    Ask the user for a password if they haven't logged in before.
    """
    conn = try_connect_delete_user_if_fail(
        sql_server_and_port,
        user_input=user_input,
        password_input=password_input,
        output=output,
    )

    return conn, conn.cursor()


def multi_connect(sql_server_and_ports):
    """Create SQL connections and cursors.
    """
    connections = dict()
    cursors = dict()

    for server_port in sql_server_and_ports:
        connections[server_port], cursors[server_port] = single_connect(server_port)

    return connections, cursors


def single_close(connection):
    connection.close()


def multi_close(connections):
    """Close the SQL connections.
    """
    for _, item in connections.items():
        single_close(item)


@contextmanager
def connect(sql_server_and_ports):
    """A controlled execution class that opens and closes multiple SQL
    connections.

    Usage example:
        import pymedphys

        with pymedphys.mosaiq.connect('msqsql') as cursor:
            do_something(cursor)
    """
    if isinstance(sql_server_and_ports, str):
        sql_server_and_ports_as_list = [sql_server_and_ports]
        return_unnested_cursor = True
    else:
        sql_server_and_ports_as_list = list(sql_server_and_ports)
        return_unnested_cursor = False

    connections, cursors = multi_connect(sql_server_and_ports_as_list)
    try:
        if return_unnested_cursor:
            cursors = cursors[sql_server_and_ports]
        yield cursors
    finally:
        multi_close(connections)


mosaiq_connect = connect
multi_mosaiq_connect = connect



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys

from pymedphys import _config
from pymedphys._vendor.patchlogging import apply_logging_patch

from .dev import dev_cli
from .dicom import dicom_cli
from .experimental import experimental_cli
from .gui import gui_cli
from .icom import icom_cli
from .trf import trf_cli
from .zenodo import zenodo_cli


class DefaultHelpParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def define_parser():
    parser = DefaultHelpParser(prog="pymedphys")
    subparsers = parser.add_subparsers()

    dicom_cli(subparsers)
    experimental_cli(subparsers)
    trf_cli(subparsers)
    dev_cli(subparsers)
    zenodo_cli(subparsers)
    icom_cli(subparsers)
    gui_cli(subparsers)

    # https://stackoverflow.com/a/20663028/3912576
    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose",
        action="store_true",
        dest="logging_verbose",
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="Print debugging statements",
        action="store_true",
        dest="logging_debug",
    )

    return parser


def get_logging_config():
    try:
        config = _config.get_config()
    except FileNotFoundError:
        return {}

    try:
        cli_config = config["cli"]
        logging_config = cli_config["logging"]
    except KeyError:
        return {}

    return logging_config


def run_logging_basic_config(args, logging_config):
    if "level" not in logging_config:
        logging_config["level"] = logging.WARNING

    # Allow command line options to override the config.toml options
    if args.logging_verbose:
        logging_config["level"] = logging.INFO
    else:
        try:
            logging_config["level"] = getattr(logging, logging_config["level"].upper())
        except AttributeError:
            pass

    # Have debug after info so that if both --verbose and --debug are
    # passed to the CLI debug will be used. Should both be passed it is
    # logged as a warning below.
    if args.logging_debug:
        logging_config["level"] = logging.DEBUG

    if "format" not in logging_config:
        logging_config["format"] = "%(asctime)s %(levelname)-8s %(message)s"

        if logging_config["level"] <= logging.DEBUG:
            logging_config["format"] += "\n    %(pathname)s#%(lineno)d"

    if "datefmt" not in logging_config:
        logging_config["datefmt"] = "%Y-%m-%d %H:%M:%S"

    logging.basicConfig(force=True, **logging_config)

    if args.logging_debug and args.logging_verbose:
        logging.warning(
            "Both --verbose and --debug were defined. Verbose mode was "
            "ignored. Debug mode has been used."
        )

    logging.info(
        "Set `logging.basicConfig` with:\n%(logging_config)s",
        {"logging_config": logging_config},
    )


def pymedphys_cli():
    _config.is_cli = True

    # This is to allow the usage of force=True within logging.basicConfig
    apply_logging_patch()

    parser = define_parser()

    args, remaining = parser.parse_known_args()
    logging_config = get_logging_config()
    run_logging_basic_config(args, logging_config)

    if hasattr(args, "func"):
        try:
            args.func(args, remaining)
            return
        except TypeError:
            pass

        parser.parse_args()
        args.func(args)

        return

    subparser_names = [
        attribute
        for attribute in dir(args)
        if not (attribute.startswith("_") or attribute.startswith("logging"))
    ]

    if not subparser_names:
        parser.print_help()
    else:
        assert len(subparser_names) == 1

        subparser_name = subparser_names[0]
        assert getattr(args, subparser_name) is None

        parser.parse_args([subparser_name, "--help"])



# Copyright 2020 University of New South Wales, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


class DicomConnectBase:
    """Base class for DICOM connect module
    """

    def __init__(self, host="127.0.0.1", port=8888, ae_title="PYMEDPHYSCONNECT"):
        """Create instance of a DicomConnect class

        Parameters
        ----------
        host : str, optional
            Host name or IP to connect to, by default "127.0.0.1"
        port : int, optional
            Port to connect to/listen on, by default 8888
        ae_title : str, optional
            AETitle of the calling/called application, by default "PYMEDPHYSCONNECT"
        """

        self.host = host
        self.port = port
        self.ae_title = ae_title

        logging.debug(
            "DicomConnect host: %s, port: %d, AE Title: %s",
            self.host,
            self.port,
            self.ae_title,
        )



# Copyright (C) 2019 Cancer Care Associates and Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

from pymedphys._utilities import patient as pmp_util_patient

from . import utility


def read_patient_name(patient_directory):
    patient_directory = pathlib.Path(patient_directory)
    patient_id = str(patient_directory.name).split("~")[1]
    demographic_file = patient_directory.joinpath(f"demographic.{patient_id}")

    if not demographic_file.exists():
        raise ValueError(
            f"Could not find demographic file at the location {str(demographic_file)}"
        )

    read_monaco_file = utility.create_read_monaco_file()

    contents = read_monaco_file(demographic_file)
    patient_name = contents.split("\n")[2]

    try:
        patient_name = pmp_util_patient.convert_patient_name(patient_name)
    except:  # pylint: disable = bare-except
        pass

    return patient_name



import ast
import importlib
import pathlib

from pymedphys._vendor import apipkg

HERE = pathlib.Path(__file__).parent

with open(HERE.joinpath("imports.py")) as f:
    imports_string = f.read()


imports_for_apipkg = {}

for node in ast.parse(imports_string).body:
    if not isinstance(node, ast.Import):
        raise ValueError("Only direct import statements are supported")

    aliases = list(node.names)
    if len(aliases) != 1:
        raise ValueError("Only one alias per import supported")

    alias = aliases[0]
    asname = alias.asname

    if asname is None:
        asname = alias.name

    imports_for_apipkg[asname] = alias.name


apipkg.initpkg(__name__, imports_for_apipkg)  # type: ignore

THIS = importlib.import_module(__name__)
IMPORTABLES = dir(THIS)

# This will never actually run, but it helps pylint know what's going on
if "numpy" not in IMPORTABLES:
    from .imports import *

    raise ValueError("This section of code should never run")



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def filter_patient_names(patient_names):
    patient_names = list(set(patient_names))

    if len(patient_names) == 1:
        patient_name = patient_names[0]
    elif len(patient_names) == 0:
        patient_name = ""
    else:
        patient_name = f"Multiple Names Found: f{', '.join(patient_names)}"

    return patient_name



from .partition import split_into_header_table
from .table import decode_rows
from .trf2pandas import header_as_dataframe


def detect_cli(args):
    detect_file_encoding(args.filepath)


def detect_file_encoding(filepath):
    with open(filepath, "rb") as file:
        trf_contents = file.read()

    trf_header_contents, trf_table_contents = split_into_header_table(trf_contents)

    header_dataframe = header_as_dataframe(trf_header_contents)
    print(header_dataframe)

    possible_groupings = search_for_possible_decoding_options(trf_table_contents)

    return possible_groupings


def search_for_possible_decoding_options(trf_table_contents):
    line_grouping_range = range(600, 800)
    linac_state_codes_column_range = range(0, 50)
    reference_state_code_keys = None

    possible_groupings = []

    for line_grouping in line_grouping_range:
        for linac_state_codes_column in linac_state_codes_column_range:
            try:
                decode_rows(
                    trf_table_contents,
                    input_line_grouping=line_grouping,
                    input_linac_state_codes_column=linac_state_codes_column,
                    reference_state_code_keys=reference_state_code_keys,
                )
                possible_groupings.append([line_grouping, linac_state_codes_column])
                print(
                    f"Line Grouping: {line_grouping}, Linac State Codes Column: {linac_state_codes_column}"
                )
            except ValueError:
                pass

    return possible_groupings



from pymedphys._imports import numpy as np

import matplotlib.pyplot as plt
from scipy import signal

from . import running_mean as rm


def minimize_junction_fieldrot(
    amplitude, peaks, peak_type, dx, profilename
):  # minimize junction for field rotations is done differently given the shape of the fields
    print("Field rotation jaw analysis...")
    # print('number of peaks=', peaks)
    amp_prev = 0
    amp_filt_prev = 0

    fig = plt.figure(figsize=(10, 6))  # create the plot

    kk = 1  # counter for figure generation
    for j in range(0, amplitude.shape[1] - 1):
        for k in range(j + 1, amplitude.shape[1]):  # looping through remaining images
            amp_base_res = signal.convolve(
                amplitude[:, j], amplitude[:, j], mode="full"
            )
            amp_base_res = signal.resample(
                amp_base_res / np.amax(amp_base_res),
                int(np.ceil(len(amp_base_res) / 2)),
            )

            amp_overlay_res = signal.convolve(
                amplitude[:, k], amplitude[:, k], mode="full"
            )
            amp_overlay_res = signal.resample(
                amp_overlay_res / np.amax(amp_overlay_res),
                int(np.ceil(len(amp_overlay_res) / 2)),
            )
            # amp_base_res = signal.savgol_filter(amplitude[:, j], 1001, 3)
            # amp_overlay_res = signal.savgol_filter(amplitude[:, k], 1001, 3)
            # peak1, _ = find_peaks(amp_base_res, prominence=0.5)
            # peak2, _ = find_peaks(amp_overlay_res, prominence=0.5)

            cumsum_prev = 1e7
            amp_base_res = amplitude[:, j]
            amp_overlay_res = amplitude[:, k]

            if peak_type[j] == 0:
                inc = -1
            else:
                inc = 1
            for i in range(0, inc * 80, inc * 1):
                # x = np.linspace(0, 0 + (len(amp_base_res) * dx), len(amplitude),
                #                 endpoint=False)  # definition of the distance axis
                amp_overlay_res_roll = np.roll(amp_overlay_res, i)

                # amplitude is the vector to analyze +-500 samples from the center
                amp_tot = (
                    amp_base_res[peaks[j] - 1000 : peaks[j] + 1000]
                    + amp_overlay_res_roll[peaks[j] - 1000 : peaks[j] + 1000]
                )  # divided by 2 to normalize
                # xsel = x[peaks[j] - 1000:peaks[j] + 1000]
                amp_filt = rm.running_mean(amp_tot, 281)

                cumsum = np.sum(np.abs(amp_tot - amp_filt))

                if (  # pylint: disable = no-else-break
                    cumsum > cumsum_prev
                ):  # then we went too far
                    ax = fig.add_subplot(amplitude.shape[1] - 1, 1, kk)

                    ax.plot(amp_prev)
                    ax.plot(amp_filt_prev)
                    if kk == 1:
                        ax.set_title(
                            "Minimization result - " + profilename, fontsize=16
                        )
                    if (
                        kk == amplitude.shape[1] - 1
                    ):  # if we reach the final plot the add the x axis label
                        ax.set_xlabel("distance [mm]")

                    ax.set_ylabel("amplitude")
                    ax.annotate(
                        "delta=" + str(abs(i - inc * 1) * dx) + " mm",
                        xy=(2, 1),
                        xycoords="axes fraction",
                        xytext=(0.35, 0.10),
                    )

                    # plt.show()

                    kk = kk + 1
                    break
                else:
                    amp_prev = amp_tot
                    amp_filt_prev = amp_filt
                    cumsum_prev = cumsum

    return fig



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

import pytest

from pymedphys._experimental.film import load_cal_scans, load_image

HERE = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(HERE, "data/spine_case")

PRESCANS_CAL_DIR = os.path.join(DATA_DIR, "DatasetA/prescans/calibration")
POSTSCANS_CAL_DIR = os.path.join(DATA_DIR, "DatasetA/postscans/calibration")

BASELINES_DIR = os.path.join(DATA_DIR, "Baselines")


@pytest.fixture
def prescans():
    filepath = Path(DATA_DIR).joinpath("DatasetA/prescans/treatment.tif")
    scans = load_cal_scans(PRESCANS_CAL_DIR)
    scans["treatment"] = load_image(filepath)

    return scans


@pytest.fixture
def postscans():
    filepath = Path(DATA_DIR).joinpath("DatasetA/postscans/treatment.tif")
    scans = load_cal_scans(POSTSCANS_CAL_DIR)
    scans["treatment"] = load_image(filepath)

    return scans



# Copyright (C) 2020 Stuart Swerdloff, Simon Biggs
# Copyright (C) 2018 Matthew Jennings, Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import logging
from copy import copy  # , deepcopy
from glob import glob
from os.path import abspath, basename, dirname
from os.path import join as pjoin

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom

from pymedphys._dicom.anonymise import strategy
from pymedphys._dicom.constants import (
    DICOM_SOP_CLASS_NAMES_MODE_PREFIXES,
    NotInBaselineError,
    get_baseline_dict_entry,
    get_baseline_keyword_vr_dict,
)

HERE = dirname(abspath(__file__))

IDENTIFYING_KEYWORDS_FILEPATH = pjoin(HERE, "identifying_keywords.json")


@functools.lru_cache()
def _get_default_identifying_keywords():
    with open(IDENTIFYING_KEYWORDS_FILEPATH) as infile:
        IDENTIFYING_KEYWORDS = json.load(infile)
    return tuple(IDENTIFYING_KEYWORDS)


def get_default_identifying_keywords():
    return list(_get_default_identifying_keywords())


def filter_identifying_keywords(keywords_to_leave_unchanged, identifying_keywords=None):
    r"""Removes DICOM keywords that the user desires to leave unchanged
    from the list of known DICOM identifying keywords and returns the
    resulting keyword list.
    """
    if identifying_keywords is None:
        keywords_filtered = get_default_identifying_keywords()
    else:
        keywords_filtered = copy(identifying_keywords)

    for keyword in keywords_to_leave_unchanged:
        try:
            keywords_filtered.remove(keyword)
        except ValueError:
            # Value not in list. TODO: Warn?
            pass

    return keywords_filtered


def non_private_tags_in_dicom_dataset(ds):
    """Return all non-private tags from a DICOM dataset.
    """

    non_private_tags = []

    for elem in ds:
        if not elem.tag.is_private and not (
            # Ignore retired Group Length elements
            elem.tag.element == 0
            and elem.tag.group > 6
        ):
            non_private_tags.append(elem.tag)
    return non_private_tags


def unknown_tags_in_dicom_dataset(ds):
    """Return all non-private tags from a DICOM dataset that do not
    exist in the PyMedPhys copy of the DICOM dictionary.
    """

    non_private_tags_in_dataset = np.array(non_private_tags_in_dicom_dataset(ds))

    are_non_private_tags_in_dict_baseline = []
    for tag in non_private_tags_in_dataset:
        try:
            get_baseline_dict_entry(tag)
            are_non_private_tags_in_dict_baseline.append(True)
        except NotInBaselineError:
            are_non_private_tags_in_dict_baseline.append(False)

    unknown_tags = list(
        non_private_tags_in_dataset[
            np.invert(np.array(are_non_private_tags_in_dict_baseline, dtype=bool))
        ]
    )

    return unknown_tags


def anonymise_tags(
    ds_anon, keywords_to_anonymise, replace_values, replacement_strategy=None
):
    """Anonymise all desired DICOM elements.
    """
    if not replace_values and replacement_strategy is not None:
        logging.warning(
            "Conflicting approach to anonymisation specified, a replacement strategy was specified in addition to a directive"
            "to eliminate values rather than replace them.  Adhering to directive to eliminate values"
        )

    for keyword in keywords_to_anonymise:
        if hasattr(ds_anon, keyword):
            if replace_values:
                if ds_anon[keyword].value in ("", None, []):
                    logging.debug(
                        "%s has value of empty list, None or empty string, no need to modify to anonymise",
                        keyword,
                    )
                    continue
                replacement_value = get_anonymous_replacement_value(
                    keyword,
                    current_value=ds_anon[keyword].value,
                    replacement_strategy=replacement_strategy,
                )
            else:
                if get_baseline_keyword_vr_dict()[keyword] in ("OB", "OW"):
                    replacement_value = (0).to_bytes(2, "little")
                else:
                    replacement_value = ""
            setattr(ds_anon, keyword, replacement_value)

    remaining_seq_only_list = [
        x for x in ds_anon if (x.VR == "SQ" and x.name not in keywords_to_anonymise)
    ]
    for seq in remaining_seq_only_list:
        for seq_item in seq.value:
            anonymise_tags(
                seq_item,
                keywords_to_anonymise,
                replace_values,
                replacement_strategy=replacement_strategy,
            )

    return ds_anon


def get_anonymous_replacement_value(
    keyword, current_value=None, replacement_strategy=None
):
    """Get an appropriate anonymisation value for a DICOM element
    based on its value representation (VR)
    Parameters
    ----------
    keyword: text string that is the pydicom name for the DICOM attribute/element

    current_value: optional, the value that is currently assigned to the element

    replacement_strategy: optional, a dispatch dictionary whose keys are the text representation
    of the VR of the element, and whose values are function references that take the current value
    of the element.

    Returns
    -------
    A value that is a suitable replacement for the element whose attributes are identified by the keyword

    TODO
    ----
    Address VR of CS to ensure DICOM conformance and if possible, interoperability
    CS typically implies a defined set of values, or in some cases, a strict enumeration of values
    and replacement with a value that is not in a defined set will often break interoperability.
    Replacement with a value that is not in an enumerated set breaks DICOM conformance.

    """
    vr = get_baseline_keyword_vr_dict()[keyword]
    if vr == "CS":
        #       An example, although this exact code breaks unit tests because
        #       the unit tests are expecting the CS hardcoded replacement string "ANON"
        #       if keyword == "PatientSex":
        #           replacement_value = "O"  # or one can replace with an empty string because PatientSex is typically type 2
        #       else:
        logging.warning(
            "Keyword %s has Value Representation CS and may require special processing to avoid breaking DICOM conformance or interoperability",
            keyword,
        )
        #   elif ...

    if replacement_strategy is None:
        replacement_strategy = strategy.ANONYMISATION_HARDCODE_DISPATCH
    try:
        replacement_value = replacement_strategy[vr](current_value)
    except KeyError:
        logging.error(
            "Unable to anonymise %s with VR %s, current value is %s",
            keyword,
            vr,
            current_value,
        )
        raise

    return replacement_value


def create_filename_from_dataset(ds, dirpath=""):
    mode_prefix = DICOM_SOP_CLASS_NAMES_MODE_PREFIXES[ds.SOPClassUID.name]
    return pjoin(dirpath, "{}.{}.dcm".format(mode_prefix, ds.SOPInstanceUID))


def label_dicom_filepath_as_anonymised(filepath):
    basename_anon = "{}_Anonymised.dcm".format(
        ".".join(basename(filepath).split(".")[:-1])
    )
    return pjoin(dirname(filepath), basename_anon)


def is_anonymised_dataset(ds, ignore_private_tags=False):
    r"""Check whether a DICOM dataset has been (fully) anonymised.

    This function specifically checks whether the dataset has been
    anonymised using a PyMedPhys anonymiser. It is very likely that it
    will return ``False`` for an anonymous dataset that was anonymised
    using a different tool.

    Parameters
    ----------
    ds : ``pydicom.dataset.Dataset``
        The DICOM dataset to check for anonymity

    ignore_private_tags : ``bool``, optional
        If set to ``False``, ``is_anonymised_dataset()`` will return
        ``False`` if any private (non-standard) DICOM tags exist in
        ``ds``. Set to ``True`` to ignore private tags when checking for
        anonymity. Do so with caution, since private tags may contain
        identifying information. Defaults to ``False``.

    Returns
    -------
    is_anonymised : ``bool``
        `True` if `ds` has been anonymised, `False` otherwise.
    """
    for elem in ds:
        if elem.keyword in get_default_identifying_keywords():
            dummy_value = get_anonymous_replacement_value(elem.keyword)
            if not elem.value in ("", [], dummy_value, None):
                if elem.VR == "DS" and np.isclose(
                    float(elem.value), float(dummy_value)
                ):
                    continue
                logging.info("%s is not considered to be anonymised", elem.name)
                return False

        elif elem.tag.is_private and not ignore_private_tags:
            logging.info(
                "%s is private and private tags are not being ignored", elem.tag
            )
            return False
        elif elem.VR == "SQ":
            contents_are_anonymous = True
            for seq in elem.value:
                contents_are_anonymous = is_anonymised_dataset(seq, ignore_private_tags)
                if not contents_are_anonymous:
                    logging.info(
                        "%s contained an element not considered to be anonymised",
                        elem.name,
                    )
                    return False

    return True


def is_anonymised_file(filepath, ignore_private_tags=False):
    r"""Check whether a DICOM file has been (fully) anonymised.

    This function specifically checks whether the DICOM file has been
    anonymised using a PyMedPhys anonymiser. It is very likely that it
    will return ``False`` for an anonymous DICOM file that was
    anonymised using a different tool.

    Parameters
    ----------
    filepath : ``str`` or ``pathlib.Path``
        The path to the DICOM file to check for anonymity.

    ignore_private_tags : ``bool``, optional
        If set to ``False``, ``is_anonymised_file()`` will return
        ``False`` if any private (non-standard) DICOM tags exist in the
        DICOM file. Set to ``True`` to ignore private tags when checking
        for anonymity. Do so with caution, since private tags may
        contain identifying information. Defaults to ``False``.

    Returns
    -------
    is_anonymised : ``bool``
        ``True`` if the DICOM dataset read from ``filepath`` has been
        anonymised, ``False`` otherwise.
    """
    ds = pydicom.dcmread(str(filepath))

    return is_anonymised_dataset(ds, ignore_private_tags=ignore_private_tags)


def is_anonymised_directory(dirpath, ignore_private_tags=False):
    r"""Check whether all DICOM files in a directory have been (fully)
    anonymised.

    This function specifically checks whether the DICOM files have been
    anonymised using a PyMedPhys anonymiser. It is very likely that it
    will return ``False`` for an anonymous DICOM file that was
    anonymised using a different tool.

    Parameters
    ----------
    dirpath : ``str`` or ``pathlib.Path``
        The path to the directory containing DICOM files to check for
        anonymity.

    ignore_private_tags : ``bool``, optional
        If set to ``False``, ``is_anonymised_directory()`` will return
        ``False`` if any private (non-standard) DICOM tags exist in any
        of the DICOM files in ``dirpath``. Set to `True` to ignore
        private tags when checking for anonymity. Do so with caution,
        since private tags may contain identifying information. Defaults
        to ``False``.

    Returns
    -------
    is_anonymised : ``bool``
        ``True`` if all of the DICOM datasets read from ``dirpath`` have
        been anonymised, ``False`` otherwise.
    """
    is_anonymised = True
    dicom_filepaths = glob(str(dirpath) + "/**/*.dcm", recursive=True)

    for dicom_filepath in dicom_filepaths:
        if not is_anonymised_file(
            dicom_filepath, ignore_private_tags=ignore_private_tags
        ):
            is_anonymised = False
            break

    return is_anonymised



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import deepcopy

import pytest

from pymedphys._dicom.create import dicom_dataset_from_dict
from pymedphys._dicom.ct.extend import (
    convert_datasets_to_deque,
    extend_datasets,
    generate_uids,
)


def make_datasets(uuids, slice_locations):
    initial_datasets = [
        dicom_dataset_from_dict(
            {
                "SOPInstanceUID": uuid,
                "InstanceNumber": str(i),
                "SliceLocation": str(float(slice_location)),
                "ImagePositionPatient": ["0.0", "0.0", str(float(slice_location))],
            }
        )
        for i, (uuid, slice_location) in enumerate(zip(uuids, slice_locations))
    ]

    return convert_datasets_to_deque(initial_datasets)


@pytest.mark.pydicom
def test_extend_datasets():
    initial_uuids = generate_uids(4)
    final_uuids = generate_uids(7)

    number_of_slices_to_add = 3
    initial_slice_locations = [1, 3, 5, 7]
    expected_slice_locations_left = [-5, -3, -1, 1, 3, 5, 7]
    expected_slice_locations_right = [1, 3, 5, 7, 9, 11, 13]

    initial_datasets = make_datasets(initial_uuids, initial_slice_locations)
    expected_datasets_left = make_datasets(final_uuids, expected_slice_locations_left)
    expected_datasets_right = make_datasets(final_uuids, expected_slice_locations_right)

    resulting_dataset_left = deepcopy(initial_datasets)
    resulting_dataset_right = deepcopy(initial_datasets)

    extend_datasets(
        resulting_dataset_left, 0, number_of_slices_to_add, uids=final_uuids
    )
    extend_datasets(
        resulting_dataset_right, -1, number_of_slices_to_add, uids=final_uuids
    )

    assert resulting_dataset_left != initial_datasets
    assert resulting_dataset_left == expected_datasets_left

    assert resulting_dataset_right != initial_datasets
    assert resulting_dataset_right == expected_datasets_right


@pytest.mark.pydicom
def test_out_of_order():
    initial_uuids = generate_uids(4)
    final_uuids = generate_uids(7)

    number_of_slices_to_add = 3
    initial_slice_locations = [7, 3, 1, 5]
    expected_slice_locations = [-5, -3, -1, 1, 3, 5, 7]

    initial_datasets = make_datasets(initial_uuids, initial_slice_locations)
    expected_datasets = make_datasets(final_uuids, expected_slice_locations)

    resulting_dataset = deepcopy(initial_datasets)

    extend_datasets(resulting_dataset, 0, number_of_slices_to_add, uids=final_uuids)

    assert resulting_dataset != initial_datasets
    assert resulting_dataset == expected_datasets



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pathlib

from pymedphys._imports import streamlit as st

from pymedphys._streamlit.utilities import config as st_config


@st.cache
def get_dicom_export_locations():
    site_directories = st_config.get_site_directories()
    dicom_export_locations = {
        site: directories["monaco"].parent.parent.joinpath("DCMXprtFile")
        for site, directories in site_directories.items()
    }

    return dicom_export_locations


@st.cache
def get_icom_live_stream_directories():
    config = st_config.get_config()
    icom_live_stream_directories = {}
    for site in config["site"]:
        icom_live_base_directory = pathlib.Path(site["export-directories"]["icom_live"])
        for linac in site["linac"]:
            icom_live_stream_directories[linac["name"]] = str(
                icom_live_base_directory.joinpath(linac["ip"])
            )

    return icom_live_stream_directories


@st.cache
def get_machine_centre_map():
    config = st_config.get_config()
    machine_centre_map = {}
    for site in config["site"]:
        for linac in site["linac"]:
            machine_centre_map[linac["name"]] = site["name"]

    return machine_centre_map


@st.cache
def get_mosaiq_details():
    config = st_config.get_config()
    mosaiq_details = {
        site["name"]: {
            "timezone": site["mosaiq"]["timezone"],
            "server": f'{site["mosaiq"]["hostname"]}:{site["mosaiq"]["port"]}',
        }
        for site in config["site"]
    }

    return mosaiq_details


@st.cache
def get_default_icom_directories():
    config = st_config.get_config()
    default_icom_directory = config["icom"]["patient_directories"]

    return default_icom_directory


@st.cache
def get_default_gamma_options():
    config = st_config.get_config()
    default_gamma_options = config["gamma"]

    return default_gamma_options


@st.cache
def get_logfile_root_dir():
    config = st_config.get_config()
    logfile_root_dir = pathlib.Path(config["trf_logfiles"]["root_directory"])

    return logfile_root_dir


@st.cache
def get_indexed_backups_directory():
    logfile_root_dir = get_logfile_root_dir()
    indexed_backups_directory = logfile_root_dir.joinpath("diagnostics/already_indexed")

    return indexed_backups_directory


@st.cache
def get_indexed_trf_directory():
    logfile_root_dir = get_logfile_root_dir()
    indexed_trf_directory = logfile_root_dir.joinpath("indexed")

    return indexed_trf_directory


def get_gamma_options(advanced_mode_local):
    default_gamma_options = get_default_gamma_options()

    if advanced_mode_local:
        st.sidebar.markdown(
            """
            # Gamma parameters
            """
        )
        result = {
            **default_gamma_options,
            **{
                "dose_percent_threshold": st.sidebar.number_input(
                    "MU Percent Threshold",
                    value=default_gamma_options["dose_percent_threshold"],
                ),
                "distance_mm_threshold": st.sidebar.number_input(
                    "Distance (mm) Threshold",
                    value=default_gamma_options["distance_mm_threshold"],
                ),
                "local_gamma": st.sidebar.checkbox(
                    "Local Gamma", default_gamma_options["local_gamma"]
                ),
                "max_gamma": st.sidebar.number_input(
                    "Max Gamma", value=default_gamma_options["max_gamma"]
                ),
            },
        }
    else:
        result = default_gamma_options

    return result



"""PyTest local plugins."""


import os

import pytest
import pytest_pylint.plugin

SKIPPING_CONFIG = {
    "slow": {
        "options": ["--run-only-slow", "--slow"],
        "help": "run the slow tests",
        "description": "mark test as slow to run",
        "skip_otherwise": True,
    },
    "cypress": {
        "options": ["--run-only-yarn", "--cypress"],
        "help": "run the cypress tests",
        "description": "mark test as using cypress",
        "skip_otherwise": True,
    },
    "pydicom": {
        "options": ["--run-only-pydicom", "--pydicom"],
        "help": "run only the tests that use pydicom",
        "description": "mark test as using pydicom",
        "skip_otherwise": False,
    },
    "pylinac": {
        "options": ["--run-only-pylinac", "--pylinac"],
        "help": "run only the tests that use pylinac",
        "description": "mark test as using pylinac",
        "skip_otherwise": False,
    },
}


# https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
def pytest_addoption(parser):
    for _, skip_item in SKIPPING_CONFIG.items():
        for option in skip_item["options"]:
            parser.addoption(
                option, action="store_true", default=False, help=skip_item["help"]
            )


def pytest_configure(config):
    for key, skip_item in SKIPPING_CONFIG.items():
        config.addinivalue_line("markers", f"{key}: {skip_item['description']}")


def pytest_collection_modifyitems(config, items):
    for key, skip_item in SKIPPING_CONFIG.items():
        this_option_set = False
        provided_option = ""

        for option in skip_item["options"]:
            if config.getoption(option):
                this_option_set = True
                provided_option = option
                break

        if not this_option_set:
            if skip_item["skip_otherwise"]:
                skip = pytest.mark.skip(
                    reason=f"need {skip_item['options'][-1]} option to run"
                )

                for item in items:
                    if key in item.keywords:
                        item.add_marker(skip)
        else:
            skip = pytest.mark.skip(reason=f"since {provided_option} was passed")
            for item in items:
                if key not in item.keywords:
                    item.add_marker(skip)

    if config.getoption("--pylint"):
        skip = pytest.mark.skip(reason="Don't run tests when running pylint")
        for item in items:
            if not isinstance(item, pytest_pylint.plugin.PyLintItem):
                item.add_marker(skip)


def pytest_ignore_collect(path, config):  # pylint: disable = unused-argument
    """return True to prevent considering this path for collection.

    This hook is consulted for all files and directories prior to
    calling more specific hooks.
    """

    relative_path = os.path.relpath(str(path), os.path.dirname(__file__))
    relative_path_list = relative_path.split(os.path.sep)

    return (
        (len(relative_path_list) > 1 and relative_path_list[0] == "examples")
        or "node_modules" in relative_path_list
        or "site-packages" in relative_path_list
        or ("_bundle" in relative_path_list and "python" in relative_path_list)
        or (
            config.getoption("--doctest-modules")
            and ("_streamlit" in relative_path_list or "tests" in relative_path_list)
        )
    )



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable = protected-access

from hypothesis import given
from hypothesis.strategies import floats

import numpy as np

import pymedphys._mocks.profiles
import pymedphys._wlutz.interppoints


@given(
    floats(-20, 20),
    floats(-20, 20),
    floats(10, 20),
    floats(10, 20),
    floats(0.5, 2),
    floats(0, 360),
)
def test_field_interp_points(x_centre, y_centre, x_edge, y_edge, penumbra, degrees):

    edge_lengths = [x_edge, y_edge]
    centre = [x_centre, y_centre]

    field = pymedphys._mocks.profiles.create_rectangular_field_function(
        centre, edge_lengths, penumbra, degrees
    )
    origin_field = pymedphys._mocks.profiles.create_rectangular_field_function(
        [0, 0], edge_lengths, penumbra, 0
    )

    (
        xx_origin_left_right,
        yy_origin_left_right,
        xx_origin_top_bot,
        yy_origin_top_bot,
    ) = pymedphys._wlutz.interppoints.define_penumbra_points_at_origin(
        edge_lengths, penumbra
    )

    points_at_origin = pymedphys._wlutz.interppoints.define_penumbra_points_at_origin(
        edge_lengths, penumbra
    )

    (
        xx_left_right,
        yy_left_right,
        xx_top_bot,
        yy_top_bot,
    ) = pymedphys._wlutz.interppoints.transform_penumbra_points(
        points_at_origin, centre, degrees
    )

    assert np.allclose(
        origin_field(xx_origin_left_right, yy_origin_left_right),
        field(xx_left_right, yy_left_right),
    )

    assert np.allclose(
        origin_field(xx_origin_top_bot, yy_origin_top_bot),
        field(xx_top_bot, yy_top_bot),
    )


@given(floats(8, 20), floats(8, 20), floats(0.5, 2))
def test_field_interp_at_origin(x_edge, y_edge, penumbra):
    centre = [0, 0]
    degrees = 0
    edge_lengths = [x_edge, y_edge]

    field = pymedphys._mocks.profiles.create_rectangular_field_function(
        centre, edge_lengths, penumbra, degrees
    )
    x_profile = pymedphys._mocks.profiles.create_profile_function(
        0, edge_lengths[0], penumbra
    )
    y_profile = pymedphys._mocks.profiles.create_profile_function(
        0, edge_lengths[1], penumbra
    )

    (
        xx_left_right,
        yy_left_right,
        xx_top_bot,
        yy_top_bot,
    ) = pymedphys._wlutz.interppoints.define_penumbra_points_at_origin(
        edge_lengths, penumbra
    )

    penumbra_range = np.linspace(-penumbra / 2, penumbra / 2, 11)

    x_penumbra_left_lookup = -edge_lengths[0] / 2 + penumbra_range
    y_penumbra_bot_lookup = -edge_lengths[1] / 2 + penumbra_range

    assert np.all(np.abs(np.diff(field(xx_left_right, yy_left_right), axis=0)) < 0.01)
    assert np.all(np.abs(np.diff(field(xx_top_bot, yy_top_bot), axis=1)) < 0.01)

    left_right = np.mean(field(xx_left_right, yy_left_right), axis=0)
    top_bot = np.mean(field(xx_top_bot, yy_top_bot), axis=1)

    assert np.allclose(left_right, left_right[::-1])
    assert np.allclose(top_bot, top_bot[::-1])

    x_profile_penumbra = np.concatenate(
        [x_profile(x_penumbra_left_lookup), x_profile(-x_penumbra_left_lookup[::-1])]
    )
    y_profile_penumbra = np.concatenate(
        [y_profile(y_penumbra_bot_lookup), y_profile(-y_penumbra_bot_lookup[::-1])]
    )

    assert np.allclose(left_right, x_profile_penumbra, rtol=0.01, atol=0.01)
    assert np.allclose(top_bot, y_profile_penumbra, rtol=0.01, atol=0.01)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np

from .interppoints import apply_transform, translate_and_rotate_transform


def transform_point(point, field_centre, field_rotation):
    transform = translate_and_rotate_transform(field_centre, field_rotation)
    bb_centre = apply_transform(*point, transform)
    bb_centre = np.array(bb_centre).tolist()

    return bb_centre


def create_centralised_field(field, centre, rotation):

    transform = translate_and_rotate_transform(centre, rotation)

    def new_field(x, y):
        x_prime, y_prime = apply_transform(x, y, transform)
        return field(x_prime, y_prime)

    return new_field



# Copyright (C) 2017 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from pymedphys._imports import numpy as np

from .core import pull_mephysto_data, pull_mephysto_item, pull_mephysto_number
from .mcc2csv import file_output


def load_single_item(filepath, index):
    distance, relative_dose, scan_curvetype, scan_depth = load_mephysto(filepath)

    return (
        distance[index],
        relative_dose[index],
        scan_curvetype[index],
        scan_depth[index],
    )


def load_mephysto(filepath, output_to_file=False, output_directory=None, sort=True):
    """Input the filepath of a mephysto .mcc file and return the data of the
    scans in four lists, distance, relative_dose, scan_curvetype, and
    scan_depth. Each respective element in these lists corresponds to an
    individual scan.
    """
    # Open the file and store the contents in file_contents
    with open(filepath) as file_pointer:
        file_contents = np.array(file_pointer.readlines())

    # Use the functions defined within mccread.py to pull the desired data
    distance, relative_dose = pull_mephysto_data(file_contents)
    scan_curvetype = pull_mephysto_item("SCAN_CURVETYPE", file_contents)
    scan_depth = pull_mephysto_number("SCAN_DEPTH", file_contents)

    # Convert python lists into numpy arrays for easier use
    distance = np.array(distance, dtype=object)
    relative_dose = np.array(relative_dose, dtype=object)
    scan_curvetype = np.array(scan_curvetype)
    scan_depth = np.array(scan_depth)

    # If the user requests to sort the data (which is default) the loaded
    # mephysto files are organised so that PDDs are first, then inplane
    # profiles, then crossplane profiles.
    if sort:
        # Find the references for where the scan type is the relevant type
        # and then use the "hstack" function to join the references together.
        sort_ref = np.hstack(
            [
                np.where(scan_curvetype == "PDD")[0],  # reference of PDDs
                np.where(scan_curvetype == "INPLANE_PROFILE")[0],  # inplane ref
                np.where(scan_curvetype == "CROSSPLANE_PROFILE")[0],  # crossplane
            ]
        )

        # Confirm that the length of sort_ref is the same as scan_curvetype.
        # This will be false if there exists an unexpected scan_curvetype.
        assert len(sort_ref) == len(scan_curvetype)

        # Apply the sorting reference to each of the relevant variables.
        distance = distance[sort_ref]
        relative_dose = relative_dose[sort_ref]
        scan_curvetype = scan_curvetype[sort_ref]
        scan_depth = scan_depth[sort_ref]

    # Output csv's if "output_to_file" is True
    if output_to_file:

        # If user didn't define an output_directory use a default one
        if output_directory is None:
            # Define output directory as a mephysto folder
            filepath_directory = os.path.dirname(filepath)
            filename = os.path.splitext(os.path.basename(filepath))[0]
            output_directory = os.path.join(filepath_directory, filename)

        # If the output directory does not exist create it
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        # Call the file_output function within csvoutput.py
        file_output(
            output_directory, distance, relative_dose, scan_curvetype, scan_depth
        )

    return distance, relative_dose, scan_curvetype, scan_depth



# Copyright (C) 2018 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from pymedphys._data import download
from pymedphys._experimental.paulking.sinogram import (
    crop,
    find_modulation_factor,
    make_histogram,
    read_bin_file,
    read_csv_file,
    unshuffle,
)


def get_sinogram_csv_path():
    return download.get_file_within_data_zip("paulking_test_data.zip", "sinogram.csv")


def test_read_csv_file():
    pat_id, results = read_csv_file(get_sinogram_csv_path())
    assert pat_id == "00000 - ANONYMOUS, PATIENT"
    num_projections = len(results)
    assert num_projections == 464
    num_leaves = len(results[0])
    assert num_leaves == 64


def test_read_bin_file():
    assert read_bin_file(
        download.get_file_within_data_zip(
            "paulking_test_data.zip", "MLC_all_test_old_800P.bin"
        )
    ).shape == (400, 64)


# convert this to a nested list


def test_crop():
    STRIP = [[0.0] * 31 + [1.0] * 2 + [0.0] * 31, [0.0] * 31 + [1.0] * 2 + [0.0] * 31]
    assert crop(STRIP) == [[1.0, 1.0], [1.0, 1.0]]


def test_unshuffle():
    unshuffled = unshuffle([[0] * 25 + [1.0] * 14 + [0] * 25] * 510)
    assert len(unshuffled) == 51  # number of angles
    assert len(unshuffled[0]) == 10  # number of projections
    assert unshuffled[0][0][0] == 0  # first leaf is closed


def test_make_histogram():
    sinogram = read_csv_file(get_sinogram_csv_path())[-1]
    assert np.allclose(make_histogram(sinogram)[0][0], [0.0, 0.1])
    assert make_histogram(sinogram)[0][1] == 25894
    # [(array([0. , 0.1]), 25894),
    #  (array([0.1, 0.2]), 0),
    #  (array([0.2, 0.3]), 11),
    #  (array([0.3, 0.4]), 3523), ...]


def test_find_modulation_factor():
    sinogram = read_csv_file(get_sinogram_csv_path())[-1]
    assert np.isclose(find_modulation_factor(sinogram), 2.762391)



# Copyright (C) 2018 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The following needs to be removed before leaving the experimental module
# pylint: skip-file

from collections import namedtuple

from pymedphys._imports import numpy as np

from scipy.interpolate import interp1d


def read_prs(file_name):
    """
    Read native Profiler data file and return dose profiles.

    Parameters
    ----------
    file_name : string
        | file name of Profiler file including path

    Returns
    -------
    Profiler : named tuple
        | Profiler.cax = float dose at central axis
        | Profiler.x = list of (x, dose) tuples
        | Profiler.y = list of (y, dose) tuples
    """

    Profiler = namedtuple("Profiler", ["cax", "x", "y"])

    with open(file_name) as profiler_file:
        for row in profiler_file.readlines():
            contents = row
            if contents[:11] == "Calibration" and "File" not in contents:
                calibs = np.array(contents.split())[1:].astype(float)
            elif contents[:5] == "Data:":
                counts = np.array(contents.split()[5:145]).astype(float)
            elif contents[:15] == "Dose Per Count:":
                dose_per_count = float(contents.split()[-1])
        assert (len(calibs)) == (len(counts)) == 140
        assert dose_per_count > 0.0
    dose = counts * dose_per_count * calibs

    y_vals = [-16.4 + 0.4 * i for i in range(83)]
    x_vals = [-11.2 + 0.4 * i for i in range(57)]

    x_prof = list(zip(y_vals, dose[:57]))
    y_prof = list(zip(x_vals, dose[57:]))

    assert np.allclose(y_prof[41][1], x_prof[28][1])
    cax_dose = y_prof[41][1]

    return Profiler(cax_dose, x_prof, y_prof)


def write_prs(
    x_prof,
    y_prof,
    file_name=None,
    version=7,
    meas_date="01/01/1970",
    meas_time="00:00:01",
    descrip="",
    instit="",
    file_cal="NONE",
    collector="Profiler2",
    coll_sn="0000000",
    coll_rev="C",
    f_ware_ver="0.0.0",
    s_ware_v="0.0.0.0",
    depth="1.0",
    nom_gain="1",
    orient="Sagittal",
    ssd="100",
    bm_mode="Pulsed",
    tray="No",
    align="None",
    interval="50",
    room="",
    mach_type="",
    mach_mod="",
    mach_num="",
    bm_modal="Photon",
    bm_energy="0",
    jaws=("5", "5", "5", "5"),
    wdg_type="None",
    wdg_ang="0",
    rate="0",
    dose="0",
    gantry="180",
    collimator="180",
    bkg_used="true",
    pulse_mode="true",
    alz_panels="0000000",
    cal_sn="0000000",
    cal_rev="A",
    temp="-100",
    dose_cal="0.001",
    d_at_cal="100.0",
    abs_cal="false",
    energy_cal="0",
    cal_date="01/01/1970",
    cal_time="00:00:01",
    cal_cmnts="",
    m_frame="false",
    updates="0",
    pulses="0",
    time="0.0",
    detectors=("83", "57", "0", "4"),
    spacing="0.4",
    concat="false",
):

    """
    Write dose profiles to a file in the native Profiler data format.

    Arguments:
        x_prof -- x profile, i.e. [(distance, dose), ...]
        y_prof -- y profile, i.e. [(distance, dose), ...]

    Keyword Argument [generally needed]:
        file_name -- output long file name, None -> return string only

    Keyword Arguments [not generally needed]:
        ** defaults to reasonable values
        ** many of these values do not affect converted profiles
        ---------
        meas_date -- string, e.g. '11/10/2018'
        meas_time -- string, e.g. '16:56:28'
        descrip   -- string description
        instit    -- string institution
        file_cal  -- long path to calibraiton file, N/A for converted
        collector -- string collector model, i.e. Profiler2
        coll_sn   -- string collector serial number, e.g. '1234567'
        coll_rev  -- string collector revision letter, e.g. 'C'
        f_ware_ver-- string firmware version, e.g. '1.2.1'
        s_ware_v  -- string software version, e.g. '1.1.0.0'
        depth     -- string measurement depth in cm
        nom_gain  -- string nominal gain, e.g. '1',
        orient    -- string orientation, e.g. 'Sagittal',
        ssd       -- string SSD in cm, e.g. '100'
        bm_mode   -- string beam mode, e.g. 'Pulsed'
        tray      -- string e.g. 'No', 'Yes'
        align     -- string, e.g. 'None', 'Light Field', 'Cross Hair'
        interval  -- string collection interval, e.g. 50, must be >0
        room      -- string room description
        mach_type -- string machine type description
        mach_mod  -- string machine model
        mach_num  -- string machine serial number
        bm_modal  -- string beam modality, e.g.'Photon', 'Electron', 'Cobalt'
        bm_energy -- string beam nominal in MV or MeV, e.g. '6'
        jaws      -- tuple of string jaw positions in cm: (l, r, t, b)
        wdg_type  -- string type, e.g. 'None', 'Static', 'Dynamic', 'Virtual'
        wdg_ang   -- string angle, e.g. '0',
        rate      -- string dose rate, e.g. '300'
        dose      -- string total dose, e.g. '100'
        gantry    -- string gantry angle in deg, e.g. '180'
        collimator-- string collimator angle in deg, e.g. '180'
        bkg_used  -- string background used, e.g. 'true'
        pulse_mode-- string pulse mode, e.g.'true'
        alz_panels-- string analyze panels, e.g. ='1015679')
        cal_sn    -- string calibration serial num, e.g. '1234567'
        cal_rev   -- string calibration revision letter, e.g. 'C',
        temp      -- string temparature, internal scale, e.g. '-98.96726775'
        dose_cal  -- string dose per count, e.g. '0.001' for 1000 cts per cGy
        d_at_cal  -- string dose used for absolute calibration, e.g. "75.40153"
        abs_cal   -- string absolute calibration, e.g. 'false', 'true'
        energy_cal-- string energy of calibration, e.g. '4'
        cal_date  -- string, e.g. '11/10/2018'
        cal_time  -- string, e.g. '16:56:28'
        cal_cmnts -- string calibration comments
        m_frame   -- string mult-frame, e.g.'false', 'true
        updates   -- string number of updates, e.g.'25',
        pulses    -- string number of pulses, e.g.'3939'
        time      -- string total measurement time in sec, e.g. '12.5'
        detectors -- tuple of string, Profiler2 -> ('83', '57', '0', '4')
        spacing   -- string detector spacing, e.g.'0.4'
        concat    -- string concatenated measuremets, e.g. 'false'


    Returns:
        The contents of the file as a list of strings.
    """

    # EXTEND PROFILE TAILS, TO ENSURE LONGER THAN DEVICE
    x_prof = [(-1000, 0)] + x_prof + [(1000, 0)]
    y_prof = [(-1000, 0)] + y_prof + [(1000, 0)]

    # X,Y DETECTOR POSITIONS FOR DEVICE
    x_vals = [-11.2 + 0.4 * i for i in range(57)]
    y_vals = [-16.4 + 0.4 * i for i in range(83)]

    # X,Y COORDINATES CORRECT, DO NOT INTERPOLATE
    if [i[0] for i in x_prof] == x_vals and [i[0] for i in y_prof] == y_vals:
        counts = (
            ["Data:", "0", "0", "0", "0"]
            + [str(int(1000 * i[1])) for i in x_prof + y_prof]
            + ["0", "0", "0", "0\n"]
        )

    else:  # INTERPOLATE X,Y COORDINATES ONTO DETECTOR POSITIONS
        interpolator = interp1d([i[0] for i in x_prof], [i[1] for i in x_prof])
        # counts_x = []
        counts_x = [float(i) for i in list(map(interpolator, x_vals))]

        interpolator = interp1d([i[0] for i in y_prof], [i[1] for i in y_prof])
        # counts_y = []
        counts_y = [float(i) for i in list(map(interpolator, y_vals))]

        # 1000 COUNTS PER CGY
        counts = (
            ["Data:", "0", "0", "0", "0"]
            + [str(int(1000 * i)) for i in counts_x + counts_y]
            + ["0", "0", "0", "0\n"]
        )

    prs_file = [
        "Version:\t {}\n".format(7),
        "Filename:\t {} \n".format("-"),
        "Date:\t {}\t Time:\t{}\n".format(meas_date, meas_time),
        "Description:\t{}\n".format(descrip),
        "Institution:\t{}\n".format(instit),
        "Calibration File:\t{}\n".format(file_cal),
        "\tProfiler Setup\n",
        "\n",
        "Collector Model:\t{}\n".format(collector),
        "Collector Serial:\t{} Revision:\t{}\n".format(coll_sn, coll_rev),
        "Firmware Version:\t{}\n".format(f_ware_ver),
        "Software Version:\t{}\n".format(s_ware_v),
        "Buildup:\t{}\tcm\tWaterEquiv\n".format(depth),
        "Nominal Gain\t{}\n".format(nom_gain),
        "Orientation:\t{}\n".format(orient),
        "SSD:\t{}\tcm\n".format(ssd),
        "Beam Mode:\t{}\n".format(bm_mode),
        "Tray Mount:\t{}\n".format(tray),
        "Alignment:\t{}\n".format(align),
        "Collection Interval:\t{}\n".format(interval),
        "\n",
        "\tMachine Data\n",
        "Room:\t{}\n".format(room),
        "Machine Type:\t{}\n".format(mach_type),
        "Machine Model:\t{}\n".format(mach_mod),
        "Machine Serial Number:\t{}\n".format(mach_num),
        "Beam Type:\t{}\tEnergy:\t{} MeV\n".format(bm_modal, bm_energy),
        "Collimator:\tLeft: {} Right: {} Top: {} Bottom: {} cm\n".format(*jaws),
        "Wedge:\t{}\tat\t{}\n".format(wdg_type, wdg_ang),
        "Rate:\t{}\tmu/Min\tDose:\t{}\n".format(rate, dose),
        "Gantry Angle:\t{} deg\tCollimator Angle:\t{} deg\n".format(gantry, collimator),
        "\n",
        "\tData Flags\n",
        "Background Used:\t{}\n".format(bkg_used),
        "Pulse Mode:\t{}\n".format(pulse_mode),
        "Analyze Panels:\t{}\n".format(alz_panels),
        "\n",
        "\tHardware Data\n",
        "Cal-Serial Number:\t{}\n".format(cal_sn),
        "Cal-Revision:\t{}\n".format(cal_rev),
        "Temperature:\t{}\n".format(temp),
        "Dose Calibration\n",
        "Dose Per Count:\t{}\n".format(dose_cal),
        "Dose:\t{}\n".format(d_at_cal),
        "Absolute Calibration:\t{}\n".format(abs_cal),
        "Energy:\t{} MV\n".format(energy_cal),
        "TimeStamp:\t{} {}\n".format(cal_date, cal_time),
        "Comments:\t{}\n".format(cal_cmnts),
        "Gain Ratios for Amp0:\t\t{}\t{}\t{}\t{}\n".format(1, 2, 4, 8),
        "Gain Ratios for Amp1:\t\t{}\t{}\t{}\t{}\n".format(1, 2, 4, 8),
        "Gain Ratios for Amp2:\t\t{}\t{}\t{}\t{}\n".format(1, 2, 4, 8),
        "Gain Ratios for Amp3:\t\t{}\t{}\t{}\t{}\n".format(1, 2, 4, 8),
        "\n",
        "Multi Frame:\t{}\n".format(m_frame),
        "# updates:\t{}\n".format(updates),
        "Total Pulses:\t{}\n".format(pulses),
        "Total Time:\t{}\n".format(0.0),
        "Detectors:\t{}\t{}\t{}\t{}\n".format(*detectors),
        "Detector Spacing:\t{}\n".format(spacing),
        "Concatenation:\t{}\n".format(concat),
        "\t".join(
            ["TYPE", "UPDATE#", "TIMETIC", "PULSES", "ERRORS"]
            + ["X" + str(i) for i in range(1, 58)]
            + ["Y" + str(i) for i in range(1, 84)]
            + ["Z0", "Z1", "Z2", "Z3\n"]
        ),
        "\t".join(["BIAS1", "", "0", "", ""] + ["0.0"] * 143 + ["0.0\n"]),
        "\t".join(["Calibration", "", "", "", ""] + ["1.0"] * 139 + ["1.0\n"]),
        "\t".join(counts),
    ]

    with open(file_name, "w") as outfile:
        for line in prs_file:
            outfile.write(line)
    return prs_file



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import lzma
import pathlib

from pymedphys._imports import pandas as pd
from pymedphys._imports import streamlit as st

from pymedphys._streamlit.apps.metersetmap import _config, _deliveries, _utilities
from pymedphys._streamlit.utilities import exceptions as _exceptions
from pymedphys._utilities import patient as utl_patient


@st.cache
def load_icom_stream(icom_path):
    with lzma.open(icom_path, "r") as f:
        contents = f.read()

    return contents


def load_icom_streams(icom_paths):
    icom_streams = []

    for icom_path in icom_paths:
        icom_streams += [load_icom_stream(icom_path)]

    return icom_streams


def icom_input_method(patient_id="", key_namespace="", advanced_mode_local=False, **_):
    icom_directories = _config.get_default_icom_directories()

    if advanced_mode_local:
        st.write("iCOM patient directories", icom_directories)

    icom_directories = [pathlib.Path(path) for path in icom_directories]

    if advanced_mode_local:
        patient_id = st.text_input(
            "Patient ID", patient_id, key=f"{key_namespace}_patient_id"
        )
        st.write(patient_id)

    icom_deliveries = []
    for path in icom_directories:
        icom_deliveries += list(path.glob(f"{patient_id}_*/*.xz"))

    icom_deliveries = sorted(icom_deliveries)

    icom_files_to_choose_from = [path.stem for path in icom_deliveries]

    timestamps = list(
        pd.to_datetime(icom_files_to_choose_from, format="%Y%m%d_%H%M%S").astype(str)
    )

    choice_path_map = dict(zip(timestamps, icom_deliveries))

    st.write(
        """
        Here you need to select the timestamps that correspond to a single
        fraction of the plan selected above. Most of the time
        you will only need to select one timestamp here, however in some
        cases you may need to select multiple timestamps.

        This can occur if for example a single fraction was delivered in separate
        beams due to either a beam interrupt, or the fraction being spread
        over multiple energies
        """
    )

    if len(timestamps) == 0:
        if patient_id != "":
            st.write(
                _exceptions.NoRecordsFound(
                    f"No iCOM delivery record found for patient ID {patient_id}"
                )
            )
        return {"patient_id": patient_id}

    if len(timestamps) == 1:
        default_timestamp = [timestamps[0]]
    else:
        default_timestamp = []

    timestamps = sorted(timestamps, reverse=True)

    try:
        selected_icom_deliveries = st.multiselect(
            "Select iCOM delivery timestamp(s)",
            timestamps,
            default=default_timestamp,
            key=f"{key_namespace}_icom_deliveries",
        )
    except st.errors.StreamlitAPIException:
        st.write(f"Default timestamp = `{default_timestamp}`")
        st.write(f"All timestamps = `{timestamps}`")
        raise

    icom_filenames = [
        path.replace(" ", "_").replace("-", "").replace(":", "")
        for path in selected_icom_deliveries
    ]

    icom_paths = []
    for selected in selected_icom_deliveries:
        icom_paths.append(choice_path_map[selected])

    if advanced_mode_local:
        st.write([str(path.resolve()) for path in icom_paths])

    patient_names = set()
    for icom_path in icom_paths:
        patient_name = str(icom_path.parent.name).split("_")[-1]
        try:
            patient_name = utl_patient.convert_patient_name_from_split(
                *patient_name.split(", ")
            )
        except:  # pylint: disable = bare-except
            pass

        patient_names.add(patient_name)

    patient_name = _utilities.filter_patient_names(patient_names)

    icom_streams = load_icom_streams(icom_paths)
    deliveries = _deliveries.cached_deliveries_loading(
        icom_streams, _deliveries.delivery_from_icom
    )

    if selected_icom_deliveries:
        identifier = f"iCOM ({icom_filenames[0]})"
    else:
        identifier = None

    if len(deliveries) == 0:
        st.write(_exceptions.InputRequired("Please select at least one iCOM delivery"))
        st.stop()

    results = {
        "site": None,
        "patient_id": patient_id,
        "patient_name": patient_name,
        "selected_icom_deliveries": selected_icom_deliveries,
        "data_paths": icom_paths,
        "identifier": identifier,
        "deliveries": deliveries,
    }

    return results



# pylint: disable = unused-import, reimported, import-error

import tkinter
import tkinter.filedialog

import attr
import black
import dateutil
import dbfread
import immutables
import keyring
import packaging
import pymssql
import requests
import timeago
import toml
import tomlkit
import tqdm
import watchdog
import watchdog.events
import watchdog.observers
import watchdog.observers.polling
import yaml

import imageio
import matplotlib
import matplotlib.patches
import matplotlib.path
import matplotlib.pyplot
import matplotlib.pyplot as plt
import matplotlib.transforms
import mpl_toolkits
import mpl_toolkits.mplot3d.art3d
import numpy
import pandas
import PIL
import scipy
import scipy.interpolate
import scipy.ndimage
import scipy.ndimage.measurements
import scipy.optimize
import scipy.signal
import scipy.special
import shapely
import shapely.affinity
import shapely.geometry
import shapely.ops
import skimage
import skimage.draw
import skimage.filters
import skimage.measure

import streamlit
import tensorflow

import libjpeg
import pydicom
import pydicom.dataset
import pydicom.filebase
import pydicom.sequence
import pydicom.uid
import pynetdicom



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
import re
import subprocess
import textwrap

from pymedphys._imports import black, tomlkit

REPO_ROOT = pathlib.Path(__file__).resolve().parent.parent.parent.parent
PYPROJECT_TOML_PATH = REPO_ROOT.joinpath("pyproject.toml")

LIBRARY_PATH = REPO_ROOT.joinpath("lib", "pymedphys")
DOCS_PATH = LIBRARY_PATH.joinpath("docs")

VERSION_PATH = LIBRARY_PATH.joinpath("_version.py")

DIST_DIR = REPO_ROOT.joinpath("dist")
SETUP_PY = REPO_ROOT.joinpath("setup.py")

REQUIREMENTS_TXT = REPO_ROOT.joinpath("requirements.txt")
REQUIREMENTS_DEV_TXT = REPO_ROOT.joinpath("requirements-dev.txt")

ROOT_PYLINT = REPO_ROOT.joinpath(".pylintrc")
LIBRARY_PYLINT = LIBRARY_PATH.joinpath(".pylintrc")

ROOT_README = REPO_ROOT.joinpath("README.rst")
DOCS_README = DOCS_PATH.joinpath("README.rst")

ROOT_CHANGELOG = REPO_ROOT.joinpath("CHANGELOG.md")
DOCS_CHANGELOG = DOCS_PATH.joinpath("CHANGELOG.md")

AUTOGEN_MESSAGE = [
    "# DO NOT EDIT THIS FILE!",
    "# This file has been autogenerated by `pymedphys dev propagate`",
]


def propagate_all(_):
    propagate_version()
    propagate_extras()
    propagate_requirements()
    propagate_pylintrc()
    propagate_readme()
    propagate_changelog()


def read_pyproject():
    with open(PYPROJECT_TOML_PATH) as f:
        pyproject_contents = tomlkit.loads(f.read())

    return pyproject_contents


def get_version_string():
    pyproject_contents = read_pyproject()
    version_string = pyproject_contents["tool"]["poetry"]["version"]

    return version_string


def propagate_version():
    version_string = get_version_string()
    version_list = re.split(r"[-\.]", version_string)

    for i, item in enumerate(version_list):
        try:
            version_list[i] = int(item)
        except ValueError:
            pass

    version_contents = textwrap.dedent(
        f"""\
        {AUTOGEN_MESSAGE[0]}
        {AUTOGEN_MESSAGE[1]}

        version_info = {version_list}
        __version__ = "{version_string}"
        """
    )

    version_contents = black.format_str(version_contents, mode=black.FileMode())

    with open(VERSION_PATH, "w") as f:
        f.write(version_contents)


def copy_file_with_autogen_message(
    original_path, target_path, comment_syntax=("# ", "")
):
    with open(original_path) as f:
        original_contents = f.read()

    new_autogen = [
        comment_syntax[0] + original_autogen[2::] + comment_syntax[1]
        for original_autogen in AUTOGEN_MESSAGE
    ]

    contents_with_autogen_warning = "\n".join(new_autogen) + "\n\n" + original_contents

    with open(target_path, "w+") as f:
        f.write(contents_with_autogen_warning)


def propagate_pylintrc():
    copy_file_with_autogen_message(ROOT_PYLINT, LIBRARY_PYLINT)


def propagate_readme():
    copy_file_with_autogen_message(ROOT_README, DOCS_README, ("..\n    ", ""))


def propagate_changelog():
    copy_file_with_autogen_message(ROOT_CHANGELOG, DOCS_CHANGELOG, ("<!-- ", " -->"))


def propagate_requirements():
    subprocess.check_call("poetry update pymedphys", shell=True)
    subprocess.check_call(
        "poetry export --without-hashes -E docs -E user -f requirements.txt --output requirements.txt",
        shell=True,
    )
    subprocess.check_call(
        "poetry export --without-hashes -E dev -f requirements.txt --output requirements-dev.txt",
        shell=True,
    )

    with open(REQUIREMENTS_TXT, "a") as f:
        f.write(".[user]\n")

    with open(REQUIREMENTS_DEV_TXT, "a") as f:
        f.write(".[dev]\n")


def propagate_extras():
    pyproject_contents = read_pyproject()

    deps = pyproject_contents["tool"]["poetry"]["dependencies"]

    extras = {}

    for key in deps:
        value = deps[key]
        comment = value.trivia.comment

        if comment.startswith("# groups"):
            split = comment.split("=")
            assert len(split) == 2
            groups = json.loads(split[-1])

            for group in groups:
                try:
                    extras[group].append(key)
                except KeyError:
                    extras[group] = [key]

    if pyproject_contents["tool"]["poetry"]["extras"] != extras:
        pyproject_contents["tool"]["poetry"]["extras"] = extras

        with open(PYPROJECT_TOML_PATH, "w") as f:
            f.write(tomlkit.dumps(pyproject_contents))



# Copyright (C) 2019 Cancer Care Associates and Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import functools

import pymedphys._utilities.filesystem


@functools.lru_cache(maxsize=1)
def create_read_monaco_file():
    def read_monaco_file(filepath):
        with pymedphys._utilities.filesystem.open_no_lock(  # pylint: disable = protected-access
            filepath, "r"
        ) as a_file:
            data = a_file.read()

        return data

    return read_monaco_file



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import pandas as pd
from pymedphys._imports import streamlit as st

import pymedphys
from pymedphys._mosaiq.delivery import NoMosaiqEntries as _NoMosaiqEntries
from pymedphys._streamlit.apps.metersetmap import _config, _deliveries, _utilities
from pymedphys._streamlit.utilities import exceptions as _exceptions
from pymedphys._streamlit.utilities import mosaiq as st_mosaiq
from pymedphys._trf.manage import index as pmp_index
from pymedphys._utilities import patient as utl_patient


@st.cache
def read_trf(trf):
    return pymedphys.read_trf(trf)


@st.cache
def _get_mosaiq_configuration(headers):
    machine_centre_map = _config.get_machine_centre_map()
    mosaiq_details = _config.get_mosaiq_details()

    centres = {machine_centre_map[machine_id] for machine_id in headers["machine"]}
    mosaiq_servers = [mosaiq_details[centre]["server"] for centre in centres]

    return machine_centre_map, mosaiq_details, mosaiq_servers


@st.cache
def get_logfile_mosaiq_info(
    headers, machine_centre_map, mosaiq_details, mosaiq_servers
):
    details = []

    cursors = {server: st_mosaiq.get_mosaiq_cursor(server) for server in mosaiq_servers}

    for _, header in headers.iterrows():
        machine_id = header["machine"]
        centre = machine_centre_map[machine_id]
        mosaiq_timezone = mosaiq_details[centre]["timezone"]
        server = mosaiq_details[centre]["server"]
        cursor = cursors[server]

        field_label = header["field_label"]
        field_name = header["field_name"]
        utc_date = header["date"]

        current_details = pmp_index.get_logfile_mosaiq_info(
            cursor, machine_id, utc_date, mosaiq_timezone, field_label, field_name
        )
        current_details = pd.Series(data=current_details)

        details.append(current_details)

    details = pd.concat(details, axis=1).T

    return details


def _attempt_patient_name_from_mosaiq(headers):
    UNKNOWN_PATIENT_NAME = "Unknown"

    st.write(
        """
        #### Corresponding Mosaiq SQL Details
        """
    )

    try:
        (
            machine_centre_map,
            mosaiq_details,
            mosaiq_servers,
        ) = _get_mosaiq_configuration(headers)
    except KeyError:
        st.warning(
            "Need Mosaiq access to determine patient name. "
            f"Patient name set to '{UNKNOWN_PATIENT_NAME}'."
        )
        patient_name = UNKNOWN_PATIENT_NAME

        return patient_name

    try:
        mosaiq_details = get_logfile_mosaiq_info(
            headers, machine_centre_map, mosaiq_details, mosaiq_servers
        )
    except _NoMosaiqEntries:
        st.warning(
            "Searched Mosaiq for an entry corresponding to this logfile. "
            "No entry was found. As such, for now, patient name has been "
            f"set to '{UNKNOWN_PATIENT_NAME}'."
        )
        patient_name = UNKNOWN_PATIENT_NAME

        return patient_name

    mosaiq_details = mosaiq_details.drop("beam_completed", axis=1)
    st.write(mosaiq_details)

    patient_names = set()
    for _, row in mosaiq_details.iterrows():
        st.write(row)
        patient_name = utl_patient.convert_patient_name_from_split(
            row["last_name"], row["first_name"]
        )
        patient_names.add(patient_name)

    patient_name = _utilities.filter_patient_names(patient_names)

    return patient_name


def trf_input_method(patient_id="", key_namespace="", **_):
    """Streamlit GUI method to facilitate TRF data provision.

    Notes
    -----
    TRF files themselves have no innate patient alignment. An option
    for TRF collection is to use the CLI tool
    ``pymedphys trf orchestrate``. This connects to the SAMBA server
    hosted on the Elekta NSS and downloads the diagnostic backup zips.
    It then takes these TRF files and queries the Mosaiq database using
    time of delivery to identify these with a patient id (Ident.Pat_ID1)
    and name.

    As such, all references to patient ID and name within this
    ``trf_input_method`` are actually a reference to their Mosaiq
    database counterparts.
    """
    FILE_UPLOAD = "File upload"
    INDEXED_TRF_SEARCH = "Search indexed TRF directory"

    import_method = st.radio(
        "TRF import method",
        [FILE_UPLOAD, INDEXED_TRF_SEARCH],
        key=f"{key_namespace}_trf_file_import_method",
    )

    if import_method == FILE_UPLOAD:
        selected_files = st.file_uploader(
            "Upload TRF files",
            key=f"{key_namespace}_trf_file_uploader",
            accept_multiple_files=True,
        )

        if not selected_files:
            return {}

        data_paths = []
        individual_identifiers = ["Uploaded TRF file(s)"]

    if import_method == INDEXED_TRF_SEARCH:
        try:
            indexed_trf_directory = _config.get_indexed_trf_directory()
        except KeyError:
            st.write(
                _exceptions.ConfigMissing(
                    "No indexed TRF directory is configured. Please use "
                    f"'{FILE_UPLOAD}' instead."
                )
            )
            return {}

        patient_id = st.text_input(
            "Patient ID", patient_id, key=f"{key_namespace}_patient_id"
        )
        st.write(patient_id)

        filepaths = list(indexed_trf_directory.glob(f"*/{patient_id}_*/*/*/*/*.trf"))

        raw_timestamps = [
            "_".join(path.parent.name.split("_")[0:2]) for path in filepaths
        ]
        timestamps = list(
            pd.to_datetime(raw_timestamps, format="%Y-%m-%d_%H%M%S").astype(str)
        )

        timestamp_filepath_map = dict(zip(timestamps, filepaths))

        timestamps = sorted(timestamps, reverse=True)

        if len(timestamps) == 0:
            if patient_id != "":
                st.write(
                    _exceptions.NoRecordsFound(
                        f"No TRF log file found for patient ID {patient_id}"
                    )
                )
            return {"patient_id": patient_id}

        if len(timestamps) == 1:
            default_timestamp = timestamps[0]
        else:
            default_timestamp = []

        selected_trf_deliveries = st.multiselect(
            "Select TRF delivery timestamp(s)",
            timestamps,
            default=default_timestamp,
            key=f"{key_namespace}_trf_deliveries",
        )

        if not selected_trf_deliveries:
            return {}

        st.write(
            """
            #### TRF filepath(s)
            """
        )

        selected_files = [
            timestamp_filepath_map[timestamp] for timestamp in selected_trf_deliveries
        ]
        st.write([str(path.resolve()) for path in selected_files])

        individual_identifiers = [
            f"{path.parent.parent.parent.parent.name} {path.parent.name}"
            for path in selected_files
        ]

        data_paths = selected_files

    st.write(
        """
        #### Log file header(s)
        """
    )

    headers = []
    tables = []
    for path_or_binary in selected_files:
        try:
            path_or_binary.seek(0)
        except AttributeError:
            pass

        header, table = read_trf(path_or_binary)
        headers.append(header)
        tables.append(table)

    headers = pd.concat(headers)
    headers.reset_index(inplace=True)
    headers.drop("index", axis=1, inplace=True)

    st.write(headers)

    deliveries = _deliveries.cached_deliveries_loading(
        tables, _deliveries.delivery_from_trf
    )

    identifier = f"TRF ({individual_identifiers[0]})"

    patient_name = _attempt_patient_name_from_mosaiq(headers)

    return {
        "site": None,
        "patient_id": patient_id,
        "patient_name": patient_name,
        "data_paths": data_paths,
        "identifier": identifier,
        "deliveries": deliveries,
    }



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# It is understood that we are stepping into 'private' API usage here
# pylint: disable = protected-access


from pymedphys._imports import streamlit as st


class SessionState:
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)


def get_session_id():
    ctx = st.report_thread.get_report_ctx()
    session_id = ctx.session_id

    return session_id


def get_session():
    session_id = get_session_id()
    return st.server.server.Server.get_current()._get_session_info(session_id).session


def initialise_session_state(**kwargs):
    session = get_session()
    if not hasattr(session, "_custom_session_state"):
        session._custom_session_state = SessionState(**kwargs)

    return session._custom_session_state



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pymedphys._imports import streamlit as st

from pymedphys._monaco import patient as mnc_patient
from pymedphys._streamlit.apps.metersetmap import _deliveries
from pymedphys._streamlit.utilities import exceptions as _exceptions
from pymedphys._streamlit.utilities import monaco as st_monaco


@st.cache
def read_monaco_patient_name(monaco_patient_directory):
    return mnc_patient.read_patient_name(monaco_patient_directory)


def monaco_input_method(
    patient_id="", key_namespace="", advanced_mode_local=False, site=None, **_
):
    (
        monaco_site,
        monaco_directory,
        patient_id,
        plan_directory,
        patient_directory,
    ) = st_monaco.monaco_patient_directory_picker(
        patient_id, key_namespace, advanced_mode_local, site
    )

    patient_name = read_monaco_patient_name(str(patient_directory))

    st.write(f"Patient Name: `{patient_name}`")

    all_tel_paths = list(plan_directory.glob("**/*tel.1"))
    all_tel_paths = sorted(all_tel_paths, key=os.path.getmtime)

    plan_names_to_choose_from = [
        str(path.relative_to(plan_directory)) for path in all_tel_paths
    ]

    if len(plan_names_to_choose_from) == 0:
        if patient_id != "":
            st.write(
                _exceptions.NoRecordsFound(
                    f"No Monaco plans found for patient ID {patient_id}"
                )
            )
        return {"patient_id": patient_id}

    st.write(
        """
        Select the Monaco plan that correspond to a patient's single fraction.
        If a patient has multiple fraction types (such as a plan with a boost)
        then these fraction types need to be analysed separately.
        """
    )

    selected_monaco_plan = st.radio(
        "Select a Monaco plan",
        plan_names_to_choose_from,
        key=f"{key_namespace}_monaco_plans",
    )

    tel_paths = []

    if selected_monaco_plan is not None:
        current_plans = list(
            monaco_directory.glob(f"*~{patient_id}/plan/{selected_monaco_plan}")
        )
        current_plans = [path.resolve() for path in current_plans]
        if len(current_plans) != 1:
            st.write("Plans found:", current_plans)
            raise ValueError("Exactly one plan should have been found")
        tel_paths += current_plans

    if advanced_mode_local:
        st.write([str(path.resolve()) for path in tel_paths])

    deliveries = _deliveries.cached_deliveries_loading(
        tel_paths, _deliveries.delivery_from_tel
    )

    if tel_paths:
        plan_names = ", ".join([path.parent.name for path in tel_paths])
        identifier = f"Monaco ({plan_names})"
    else:
        identifier = None

    if len(deliveries) == 1 and len(deliveries[0].mu) == 0:
        st.write(
            _exceptions.NoControlPointsFound(
                "This is likely due to an as of yet unsupported "
                "Monaco file format. At this point in time 3DCRT "
                "is not yet supported for reading directly from "
                "Monaco. DICOM is though, please export the plan "
                "to RT DICOM and import the data that way."
            )
        )

    results = {
        "site": monaco_site,
        "patient_id": patient_id,
        "patient_name": patient_name,
        "selected_monaco_plan": selected_monaco_plan,
        "data_paths": tel_paths,
        "identifier": identifier,
        "deliveries": deliveries,
    }

    return results
